/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
/*                          _______      ______    ______          __      */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~    / ____(_)___ / ___ /)  / ____/___  ____/ /__    */
/*     [fsWinApp]         / /_  / / __ \\__ \|/  / /   / __ \/ __  / _ \   */
/*       rev. 4          / __/ / / /_/ /__/ /   / /___/ /_/ / /_/ /  __/   */
/*    9th Feb 2007      /_/   /_/ ,___/____/    \____/\____/\__,_/\___/    */
/*     [ ] stable              /_/ (c) 2006-7 Filip STOKLAS (FipS)         */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~       http://HOLE.4FipS.com/fips_code.php          */
/*                                                                         */
/* This code is free for personal and commercial use. You may redistribute */
/* it by any means. If you use the code for your own projects please give  */
/* me credit. Please send a bug report. Don't alter or remove this header! */
/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
#ifndef FS_RAP_FSWINAPP_H
#define FS_RAP_FSWINAPP_H
//---------------------------------------------------------------------------
#if defined(FS_INCLUDE_USERDEFS) // inject user definition file on request
#   include <fs/fsUserDefs.h>
#endif 
//---------------------------------------------------------------------------
namespace fs { namespace rap {
//---------------------------------------------------------------------------
//! Single-window Application.
/*! The class represents a <b>Windows application</b> and its <b>main
    window</b>. Such a skeleton can be then <b>extended through custom
    client classes</b> offering particular functionality. A client class
    implements WinApp_t::EvRcvr_i <b>interface</b> to define <b>responses
    to individual events</b>. It's especially useful for <b>rapid
    prototyping </b> as only a simple client class, encapsulating requested
    behaviour, has to be implemented. From the user's point of view, the
    <b>entry-point</b> of the application is the <b>WinAppInit</b> function
    in the <b>fs::rap namespace</b>, which has to be provided by the user.
    Typically, the function performs two actions: it <b>modifies default
    configuration</b> and <b>registers event receivers</b> (clients). Note
    that clients can be layered in this way.
    
    \code
    namespace fs { namespace rap {
    bool WinAppInit(WinApp_t &App, WinApp_t::InitCfg_t &Cfg)
    {
        // modify default configuration
        Cfg.m_nClW = 512;
        Cfg.m_nClH = 512;

        // register event receivers (clients)
        static test::Client_t Client;
        App.RegEvRcvr(&Client);

        return true;
    }
    }} // namespace fs::rap
    \endcode
    
    \par For more information see:
    
    - InitCfg_t for changing <b>initial configuration</b>.
    - EvRcvr_i for <b>event system</b> details.
*/
class WinApp_t
{
 public:
 
    //! Init event, empty.
    struct InitEv_t {};
    //! Quit event, empty.
    struct QuitEv_t {};
    //! Run-time task event, empty.
    struct RunEv_t {};
    //! Redraw request event, empty.
    struct DrawEv_t {};
    struct SizeEv_t;
    struct KeyEv_t;
    struct MouseEv_t;
    struct EvRcvr_i;
    
    //! Checks the internal state / correct construction.
    bool IsValid() const;
        
    //! Returns the command-line string passed to the app. by the system.
    const char * GetCmdLine() const;
    //! Sets the title of the application window.
    void SetTitle(const char *pszTitle);
    
    //! Sends the 'Redraw' request.
    void Redraw();
    //! Sends the 'Fullscreen' request.
    void Fscrn(bool bOn);
    //! Sends the 'Quit' request.
    void Quit();
    
    //! Registers a new event event receiver (client).
    void RegEvRcvr(EvRcvr_i *pEvRcvr);
    
    //! Returns a system handle of the application window (should not be 0).
    void * GetWinHdle() const;
            
    //! Event receiver (client) interface.
    struct EvRcvr_i
    {
        virtual ~EvRcvr_i() {}
        virtual void OnEvent(const InitEv_t &, WinApp_t &) {}
        virtual void OnEvent(const QuitEv_t &, WinApp_t &) {}
        virtual void OnEvent(const RunEv_t &, WinApp_t &) {}
        virtual void OnEvent(const DrawEv_t &, WinApp_t &) {}
        virtual void OnEvent(const SizeEv_t &, WinApp_t &) {}
        virtual void OnEvent(const KeyEv_t &, WinApp_t &) {}
        virtual void OnEvent(const MouseEv_t &, WinApp_t &) {}
    };
 
    //! Initial configuration.
    struct InitCfg_t
    {
        const char *m_pszTitle; //!< Text in the caption bar.
        int m_nClW; //!< Width of the client area.
        int m_nClH; //!< Height of the client area.
        bool m_bSzbl; //!< Allow window resizing.
        bool m_bClrBgr; //!< Clear the background before redrawing.
        bool m_bRtm; //!< Enable real-time mode (generates RunEv_t event).
        bool m_bFscrn; //!< Launch in the full-screen mode.
        int m_nFscrnW; //!< Full-screen width.
        int m_nFscrnH; //!< Full-screen height.
                
        InitCfg_t():
        m_pszTitle("FipS' Rapid Window Application"),
        m_nClW(640),
        m_nClH(480),
        m_bSzbl(true),
        m_bClrBgr(true),
        m_bRtm(false),
        m_bFscrn(false),
        m_nFscrnW(-1),
        m_nFscrnH(-1)
        {}
    };

    //! Resize event.
    struct SizeEv_t
    {
        int m_nClW, m_nClH;

        SizeEv_t(int nClW, int nClH):
        m_nClW(nClW), m_nClH(nClH) {}

     private:
     
        SizeEv_t();
    };
 
    //! Key / mouse button event.
    struct KeyEv_t
    {
        //! Special key codes.
        enum Key_e
        {
            KEY_LMBTN = -1, //!< Left mouse button.
            KEY_RMBTN = -2, //!< Right mouse button.
            KEY_MMBTN = -3, //!< Middle mouse button.
        };
    
        int m_nKeyCode;
        bool m_bDown;
        int m_nCurX, m_nCurY;
    
        KeyEv_t(int nKeyCode, bool bDown, int nCurX, int nCurY):
        m_nKeyCode(nKeyCode),
        m_bDown(bDown),
        m_nCurX(nCurX), m_nCurY(nCurY)
        {}
        
     private:
     
        KeyEv_t();
    };
    
    //! Mouse position event.
    struct MouseEv_t
    {
        int m_nCurX, m_nCurY;

        MouseEv_t(int nCurX, int nCurY):
        m_nCurX(nCurX), m_nCurY(nCurY) {}

     private:
     
        MouseEv_t();
    };
    
 private:
 
    friend WinApp_t & WinAppInst();
 
    WinApp_t() {}
    WinApp_t(const WinApp_t &); //!< No copy.
    WinApp_t & operator = (const WinApp_t &); //!< No assignment.
};
//---------------------------------------------------------------------------
extern bool WinAppInit(WinApp_t &App, WinApp_t::InitCfg_t &Cfg);
//---------------------------------------------------------------------------
}} // namespace fs::rap
//---------------------------------------------------------------------------
#endif // FS_RAP_FSWINAPP_H