/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
/*                          _______      ______    ______          __      */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~    / ____(_)___ / ___ /)  / ____/___  ____/ /__    */
/*     [fsRle8Enc]        / /_  / / __ \\__ \|/  / /   / __ \/ __  / _ \   */
/*       rev. 6          / __/ / / /_/ /__/ /   / /___/ /_/ / /_/ /  __/   */
/*    6th Feb 2007      /_/   /_/ ,___/____/    \____/\____/\__,_/\___/    */
/*     [x] stable              /_/ (c) 2006-7 Filip STOKLAS (FipS)         */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~       http://HOLE.4FipS.com/fips_code.php          */
/*                                                                         */
/* This code is free for personal and commercial use. You may redistribute */
/* it by any means. If you use the code for your own projects please give  */
/* me credit. Please send a bug report. Don't alter or remove this header! */
/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
#include <fs/ras/fsRle8Enc.h>

// >>> COMMON SUPPORT BLOCK (rev. 6)
#if defined(FS_HAS_FSASSERT)
#   include <fs/sys/fsAssert.h>
#else
#   include <cassert>
#   if !defined(FS_ASSERT)
#       define FS_ASSERT(exp) assert(exp)
#   endif
#   if !defined(FS_ASSERT_MSG)
#       define FS_ASSERT_MSG(exp, msg) assert(exp && msg)
#   endif
#   if !defined(FS_VERIFY)
#       define FS_VERIFY(exp) \
        { bool bExp = !!(exp); assert(bExp && #exp); bExp; }
#   endif
#   if !defined(FS_VERIFY_MSG)
#       define FS_VERIFY_MSG(exp, msg) \
        { bool bExp = !!(exp); assert(bExp && #exp && msg); bExp; }
#   endif
#   if !defined(FS_STATIC_ASSERT)
#       define FS_STATIC_ASSERT(exp) { char error[(exp) ? 1 : 0]; error; }
#   endif
#endif
#if !defined(FS_VERIFY_RETURN)
#   define FS_VERIFY_RETURN(exp) \
    if(!(exp)) { FS_ASSERT_MSG(0, #exp "<return>"); return; }
#endif
#if !defined(FS_VERIFY_RETURN_VAL)
#   define FS_VERIFY_RETURN_VAL(exp, ret) \
    if(!(exp)) { FS_ASSERT_MSG(0, #exp "<returns:>" #ret); return (ret); }
#endif
// <<< COMMON SUPPORT BLOCK (rev. 6)

#include <vector>
#include <algorithm>

// provide detailed memory tracking information
#if defined(FS_HAS_FSMEMMGR)
#   include <fs/sys/fsMemMgr.h>
#endif
//---------------------------------------------------------------------------
using namespace fs::ras;
//---------------------------------------------------------------------------
//--- ANONYMOUS NAMESPACE >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
//---------------------------------------------------------------------------
namespace {
//---------------------------------------------------------------------------
//--- CLASS Rle8EncImpl_t >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
//---------------------------------------------------------------------------
class Rle8EncImpl_t
{
public:
    
    Rle8EncImpl_t(const Rle8Enc_t::PxFmt8888_t *pSrc, int nW, int nH);
    
    bool IsValid() const;
    
    bool GetSize(int &nW, int &nH) const;
    bool GetRle(const unsigned char *&pRle, int &nNum) const;
    bool GetPal(const Rle8Enc_t::PxFmt8888_t *&pPal, int &nNum) const;
    bool GetOff(const Rle8Enc_t::uint32 *&pOff, int &nNum)
     const;

private:
    
    Rle8EncImpl_t();
    Rle8EncImpl_t(const Rle8EncImpl_t &);
    Rle8EncImpl_t & operator = (const Rle8EncImpl_t &);
    
    void Shrink();
    
    unsigned char GetClrIdx(const Rle8Enc_t::PxFmt8888_t &Clr) const;
    const Rle8Enc_t::PxFmt8888_t & GetClr(int nIdx) const;
    
    //---
    
    int m_nW, m_nH;
    
    typedef std::vector<unsigned char> TRle;
    TRle m_Rle;
    
    typedef std::vector<Rle8Enc_t::PxFmt8888_t> TPal;
    mutable TPal m_Pal;
            
    typedef std::vector<Rle8Enc_t::uint32> TOff;
    TOff m_Off;
};
//---------------------------------------------------------------------------
Rle8EncImpl_t::Rle8EncImpl_t(const Rle8Enc_t::PxFmt8888_t *pSrc, int nW,
 int nH):
m_nW(nW),
m_nH(nH)
{
    FS_VERIFY_RETURN(pSrc);
                
    for(int y = 0; y < nH; ++y)
    {
        m_Off.push_back(static_cast<Rle8Enc_t::uint32>(m_Rle.size()));
    
        unsigned char uN = 1;
        for(int x = 0; x < nW; ++x)
        {
            bool bTr = pSrc->a == 0; // current
            bool bTrNx = x < nW - 1 ? (pSrc + 1)->a == 0 : !bTr; // next
            
            // end of coherent pixel sequence or the right edge
            if(bTr != bTrNx || uN >= 127)
            {
                // write number of pixels + transparency flag
                m_Rle.push_back(uN << 1 | (bTr ? 0 : 1));
                
                // write opaque pixel sequence
                if(!bTr)
                {
                    const Rle8Enc_t::PxFmt8888_t *pSrcB = pSrc - uN + 1;
                    while(pSrcB <= pSrc)
                    {
                        unsigned char uClrIdx = GetClrIdx(*pSrcB);
                        m_Rle.push_back(uClrIdx);
                        ++pSrcB;
                    }
                }
                
                uN = 0;
            }

            // go to the next pixel
            ++uN;
            ++pSrc;
        }
    }
    
    Shrink();
}
//---------------------------------------------------------------------------
bool Rle8EncImpl_t::IsValid() const
{
    return !m_Rle.empty();
}
//---------------------------------------------------------------------------
bool Rle8EncImpl_t::GetSize(int &nW, int &nH) const
{
    FS_VERIFY_RETURN_VAL(IsValid(), false);

    nW = m_nW;
    nH = m_nH;
    
    return true;
}
//---------------------------------------------------------------------------
bool Rle8EncImpl_t::GetRle(const unsigned char *&pRle, int &nNum)
 const
{
    FS_VERIFY_RETURN_VAL(IsValid(), false);

    nNum = static_cast<int>(m_Rle.size());
    pRle = nNum ? &m_Rle[0] : 0;
        
    return true;
}
//---------------------------------------------------------------------------
bool Rle8EncImpl_t::GetPal(const Rle8Enc_t::PxFmt8888_t *&pPal, int &nNum)
 const
{
    FS_VERIFY_RETURN_VAL(IsValid(), false);

    nNum = static_cast<int>(m_Pal.size());
    pPal = nNum ? &m_Pal[0] : 0;
    
    return true;
}
//---------------------------------------------------------------------------
bool Rle8EncImpl_t::GetOff(const Rle8Enc_t::uint32 *&pOff, int &nNum)
 const
{
    FS_VERIFY_RETURN_VAL(IsValid(), false);
    FS_STATIC_ASSERT(sizeof(Rle8Enc_t::uint32) == 4);
    
    nNum = static_cast<Rle8Enc_t::uint32>(m_Off.size());
    pOff = nNum ? &m_Off[0] : 0;
        
    return true;
}
//---------------------------------------------------------------------------
void Rle8EncImpl_t::Shrink()
{
    TRle(m_Rle).swap(m_Rle);
    TPal(m_Pal).swap(m_Pal);
    TOff(m_Off).swap(m_Off);
}
//---------------------------------------------------------------------------
//! \todo Check the color limit.
unsigned char Rle8EncImpl_t::GetClrIdx(const Rle8Enc_t::PxFmt8888_t &Clr) const
{
    TPal::const_iterator it = std::find(m_Pal.begin(), m_Pal.end(), Clr);
    
    if(it == m_Pal.end())
    {
        m_Pal.push_back(Clr);
        return static_cast<unsigned char>(m_Pal.size() - 1);
    }
    else
        return static_cast<unsigned char>(it - m_Pal.begin());
}
//---------------------------------------------------------------------------
const Rle8Enc_t::PxFmt8888_t & Rle8EncImpl_t::GetClr(int nIdx) const
{
    FS_ASSERT(nIdx >= 0 && nIdx < static_cast<int>(m_Pal.size()));
    return m_Pal[nIdx];
}
//---------------------------------------------------------------------------
} // anonymous namespace
//---------------------------------------------------------------------------
//--- CLASS Rle8Enc_t >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
//---------------------------------------------------------------------------
Rle8Enc_t::Rle8Enc_t(const PxFmt8888_t *pSrc, int nW, int nH):
m_pImpl(new Rle8EncImpl_t(pSrc, nW, nH))
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
}
//---------------------------------------------------------------------------
Rle8Enc_t::~Rle8Enc_t()
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
    delete static_cast<Rle8EncImpl_t *>(m_pImpl);
}
//---------------------------------------------------------------------------
bool Rle8Enc_t::IsValid() const
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
    return static_cast<Rle8EncImpl_t *>(m_pImpl)->IsValid();
}
//---------------------------------------------------------------------------
bool Rle8Enc_t::GetSize(int &nW, int &nH) const
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
    return static_cast<Rle8EncImpl_t *>(m_pImpl)->GetSize(nW, nH);
}
//---------------------------------------------------------------------------
bool Rle8Enc_t::GetRle(const unsigned char *&pRle, int &nNum) const
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
    return static_cast<Rle8EncImpl_t *>(m_pImpl)->GetRle(pRle, nNum);
}
//---------------------------------------------------------------------------
bool Rle8Enc_t::GetPal(const PxFmt8888_t *&pPal, int &nNum)
 const
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
    return static_cast<Rle8EncImpl_t *>(m_pImpl)->GetPal(pPal, nNum);
}
//---------------------------------------------------------------------------
bool Rle8Enc_t::GetOff(const uint32 *&pOff, int &nNum) const
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
    return static_cast<Rle8EncImpl_t *>(m_pImpl)->GetOff(pOff, nNum);
}
//---------------------------------------------------------------------------