/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
/*                          _______      ______    ______          __      */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~    / ____(_)___ / ___ /)  / ____/___  ____/ /__    */
/*     [fsTgaLdr]         / /_  / / __ \\__ \|/  / /   / __ \/ __  / _ \   */
/*       rev. 4          / __/ / / /_/ /__/ /   / /___/ /_/ / /_/ /  __/   */
/*    5th Feb 2007      /_/   /_/ ,___/____/    \____/\____/\__,_/\___/    */
/*     [x] stable              /_/ (c) 2006-7 Filip STOKLAS (FipS)         */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~       http://HOLE.4FipS.com/fips_code.php          */
/*                                                                         */
/* This code is free for personal and commercial use. You may redistribute */
/* it by any means. If you use the code for your own projects please give  */
/* me credit. Please send a bug report. Don't alter or remove this header! */
/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
#include <fs/ras/fsTgaLdr.h>

// >>> COMMON SUPPORT BLOCK (rev. 6)
#if defined(FS_HAS_FSASSERT)
#   include <fs/sys/fsAssert.h>
#else
#   include <cassert>
#   if !defined(FS_ASSERT)
#       define FS_ASSERT(exp) assert(exp)
#   endif
#   if !defined(FS_ASSERT_MSG)
#       define FS_ASSERT_MSG(exp, msg) assert(exp && msg)
#   endif
#   if !defined(FS_VERIFY)
#       define FS_VERIFY(exp) \
        { bool bExp = !!(exp); assert(bExp && #exp); bExp; }
#   endif
#   if !defined(FS_VERIFY_MSG)
#       define FS_VERIFY_MSG(exp, msg) \
        { bool bExp = !!(exp); assert(bExp && #exp && msg); bExp; }
#   endif
#   if !defined(FS_STATIC_ASSERT)
#       define FS_STATIC_ASSERT(exp) { char error[(exp) ? 1 : 0]; error; }
#   endif
#endif
#if !defined(FS_VERIFY_RETURN)
#   define FS_VERIFY_RETURN(exp) \
    if(!(exp)) { FS_ASSERT_MSG(0, #exp "<return>"); return; }
#endif
#if !defined(FS_VERIFY_RETURN_VAL)
#   define FS_VERIFY_RETURN_VAL(exp, ret) \
    if(!(exp)) { FS_ASSERT_MSG(0, #exp "<returns:>" #ret); return (ret); }
#endif
// <<< COMMON SUPPORT BLOCK (rev. 6)

#include <stdio.h>
#include <vector>

// provide detailed memory tracking information
#if defined(FS_HAS_FSMEMMGR)
#   include <fs/sys/fsMemMgr.h>
#endif
//---------------------------------------------------------------------------
#if (_MSC_VER >= 1400) // MSVC8
#   pragma warning(disable : 4996) // function was marked as deprecated
#endif
//---------------------------------------------------------------------------
using namespace fs::ras;
//---------------------------------------------------------------------------
//--- ANONYMOUS NAMESPACE >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
//---------------------------------------------------------------------------
namespace {
//---------------------------------------------------------------------------
template<typename T>
bool ReadType(T &Dest, FILE *pIn)
{
    return ::fread(&Dest, sizeof(T), 1, pIn) == 1;
}
//---------------------------------------------------------------------------
struct TgaHdr_t
{
    typedef unsigned char uint8;
    typedef unsigned short uint16;

    uint8  m_uTagLen;
    uint8  m_uNoPal;
    uint8  m_uImgFmt;
    uint16 m_uPalStart;
    uint16 m_uPalLen;
    uint8  m_uPalBits;
    uint16 m_uXoff;
    uint16 m_uYoff;
    uint16 m_uW;
    uint16 m_uH;
    uint8  m_uPxBits;
    uint8  m_uImgDsc;
    
    bool Read(FILE *pIn)
    {
        FS_STATIC_ASSERT(sizeof(uint8) == 1);
        FS_STATIC_ASSERT(sizeof(uint16) == 2);
    
        if(!ReadType(m_uTagLen, pIn)) return false;
        if(!ReadType(m_uNoPal, pIn)) return false;
        if(!ReadType(m_uImgFmt, pIn)) return false;
        if(!ReadType(m_uPalStart, pIn)) return false;
        if(!ReadType(m_uPalLen, pIn)) return false;
        if(!ReadType(m_uPalBits, pIn)) return false;
        if(!ReadType(m_uXoff, pIn)) return false;
        if(!ReadType(m_uYoff, pIn)) return false;
        if(!ReadType(m_uW, pIn)) return false;
        if(!ReadType(m_uH, pIn)) return false;
        if(!ReadType(m_uPxBits, pIn)) return false;
        if(!ReadType(m_uImgDsc, pIn)) return false;
    
        return true;
    }
};
//---------------------------------------------------------------------------
//--- CLASS TgaLdrImpl_t >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
//---------------------------------------------------------------------------
class TgaLdrImpl_t
{
public:
    
    explicit TgaLdrImpl_t(const char *pszFile);
    bool IsValid() const;
    
    int GetW() const;
    int GetH() const;
    const TgaLdr_t::PxFmt888_t * GetData888() const;
    const TgaLdr_t::PxFmt8888_t * GetData8888() const;

private:

    TgaLdrImpl_t(); //!< No def ctor.
    TgaLdrImpl_t(const TgaLdrImpl_t &); //!< No copy.
    TgaLdrImpl_t & operator = (const TgaLdrImpl_t &); //!< No assignment.
    
    //---
    
    int m_nW, m_nH;
    int m_nPxBits;
    typedef std::vector<TgaHdr_t::uint8> Data_t;
    Data_t m_Data;
};
//---------------------------------------------------------------------------
TgaLdrImpl_t::TgaLdrImpl_t(const char *pszFile):
m_nW(0),
m_nH(0),
m_nPxBits(0)
{
    FS_VERIFY_RETURN(pszFile);

    //--- READ IMAGE DATA ---
    
    // open the file for reading
    FILE *pIn = ::fopen(pszFile, "rb");
    if(!pIn)
        return; // couldn't open the file
        
    // read TGA header
    TgaHdr_t Hdr;
    if(!Hdr.Read(pIn))
    {
        ::fclose(pIn);
        return; // couldn't read the header
    }

    // check image data format
    if(Hdr.m_uImgFmt != 2) // (A)RGB, uncompressed
    {
        ::fclose(pIn);
        return; // unsupported data format
    }
    
    // check pixel depth
    if(Hdr.m_uPxBits != 24 && Hdr.m_uPxBits != 32)
    {
        ::fclose(pIn);
        return; // unsupported pixel depth
    }
            
    // read image data itself    
    size_t uDataSize = Hdr.m_uW * Hdr.m_uH * Hdr.m_uPxBits / 8;
    Data_t Data;
    Data.resize(uDataSize);
    
    if(::fread(&Data[0], uDataSize, 1, pIn) != 1)
    {
        ::fclose(pIn);
        return; // IO error
    }
    
    // close the file
    ::fclose(pIn);
    
    //--- REARRANGE IMAGE DATA ---

    if(Hdr.m_uPxBits == 24) // BGR
    {
        // swap channels: BGR => RGB
        for(size_t i = 0, n = Hdr.m_uW * Hdr.m_uH * 3; i < n; i += 3)
        {
            TgaHdr_t::uint8 tmp = Data[i];
            Data[i]     = Data[i + 2];
            Data[i + 2] = tmp;
        }
    }
    else if(Hdr.m_uPxBits == 32) // BGRA
    {
        // swap channels: BGRA => RGBA
        for(size_t i = 0, n = Hdr.m_uW * Hdr.m_uH * 4; i < n; i += 4)
        {
            TgaHdr_t::uint8 tmp = Data[i];
            Data[i]     = Data[i + 2];
            Data[i + 2] = tmp;
        }
    }
    
    if((Hdr.m_uImgDsc & 1 << 5) == 0) // LB origin
    {
        // flip horizontally
        const int uPxBytes = Hdr.m_uPxBits / 8;
        for(int y = 0, yn = Hdr.m_uH / 2; y < yn; ++y)
        {
            TgaHdr_t::uint8 *pSrc1 = &Data[0] + Hdr.m_uW * uPxBytes * y;
            TgaHdr_t::uint8 *pSrc2 = &Data[0] + Hdr.m_uW * uPxBytes *
             (Hdr.m_uH - y - 1);
                                    
            for(int x = 0, xn = Hdr.m_uW * uPxBytes; x < xn; ++x)
            {
                TgaHdr_t::uint8 tmp = *pSrc1;
                *pSrc1 = *pSrc2;
                *pSrc2 = tmp;
                ++pSrc1;
                ++pSrc2;
            }
        }
    }
    
    //--- TRANSFER THE STATE ---
    
    m_nW = Hdr.m_uW;
    m_nH = Hdr.m_uH;
    m_nPxBits = Hdr.m_uPxBits;
    m_Data.swap(Data);
}
//---------------------------------------------------------------------------
bool TgaLdrImpl_t::IsValid() const
{
    return !m_Data.empty();
}
//---------------------------------------------------------------------------
int TgaLdrImpl_t::GetW() const
{
    FS_VERIFY_RETURN_VAL(IsValid(), 0);
    return m_nW;
}
//---------------------------------------------------------------------------
int TgaLdrImpl_t::GetH() const
{
    FS_VERIFY_RETURN_VAL(IsValid(), 0);
    return m_nH;
}
//---------------------------------------------------------------------------
const TgaLdr_t::PxFmt888_t * TgaLdrImpl_t::GetData888() const
{
    FS_VERIFY_RETURN_VAL(IsValid(), 0);
    FS_STATIC_ASSERT(sizeof(TgaLdr_t::PxFmt888_t) == 3);
    
    return m_nPxBits == 24 ?
     reinterpret_cast<const TgaLdr_t::PxFmt888_t *>(&m_Data[0]) : 0;
}
//---------------------------------------------------------------------------
const TgaLdr_t::PxFmt8888_t * TgaLdrImpl_t::GetData8888() const
{
    FS_VERIFY_RETURN_VAL(IsValid(), 0);
    FS_STATIC_ASSERT(sizeof(TgaLdr_t::PxFmt8888_t) == 4);

    return m_nPxBits == 32 ?
     reinterpret_cast<const TgaLdr_t::PxFmt8888_t *>(&m_Data[0]) : 0;
}
//---------------------------------------------------------------------------
} // anonymous namespace
//---------------------------------------------------------------------------
//--- CLASS TgaLdr_t >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
//---------------------------------------------------------------------------
TgaLdr_t::TgaLdr_t(const char *pszFile):
m_pImpl(new TgaLdrImpl_t(pszFile))
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
}
//---------------------------------------------------------------------------
TgaLdr_t::~TgaLdr_t()
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
    delete static_cast<TgaLdrImpl_t *>(m_pImpl);
}
//---------------------------------------------------------------------------
bool TgaLdr_t::IsValid() const
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
    return static_cast<TgaLdrImpl_t *>(m_pImpl)->IsValid();
}
//---------------------------------------------------------------------------
int TgaLdr_t::GetW() const
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
    return static_cast<TgaLdrImpl_t *>(m_pImpl)->GetW();
}
//---------------------------------------------------------------------------
int TgaLdr_t::GetH() const
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
    return static_cast<TgaLdrImpl_t *>(m_pImpl)->GetH();
}
//---------------------------------------------------------------------------
const TgaLdr_t::PxFmt888_t * TgaLdr_t::GetData888() const
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
    return static_cast<TgaLdrImpl_t *>(m_pImpl)->GetData888();
}
//---------------------------------------------------------------------------
const TgaLdr_t::PxFmt8888_t * TgaLdr_t::GetData8888() const
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
    return static_cast<TgaLdrImpl_t *>(m_pImpl)->GetData8888();
}
//---------------------------------------------------------------------------