/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
/*                          _______      ______    ______          __      */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~    / ____(_)___ / ___ /)  / ____/___  ____/ /__    */
/*      [fsAssert]        / /_  / / __ \\__ \|/  / /   / __ \/ __  / _ \   */
/*       rev. 32         / __/ / / /_/ /__/ /   / /___/ /_/ / /_/ /  __/   */
/*    25th Jan 2007     /_/   /_/ ,___/____/    \____/\____/\__,_/\___/    */
/*     [x] stable              /_/ (c) 2004-7 Filip STOKLAS (FipS)         */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~       http://HOLE.4FipS.com/fips_code.php          */
/*                                                                         */
/* This code is free for personal and commercial use. You may redistribute */
/* it by any means. If you use the code for your own projects please give  */
/* me credit. Please send a bug report. Don't alter or remove this header! */
/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
/*! \file fsAssert.h
    \brief Various assert macros.
    
    fsAssert.h brings extended versions of the default <b>assert</b> macro,
    which is typically used to <b>identify logical errors</b> in the
    <b>debug mode</b> during the development. If you are not familiar with
    the concept, please check a relevant documentation.

    Following macros are provided:
    
    - FS_ASSERT(exp),
    - FS_ASSERT_MSG(exp, msg)
    - FS_ASSERT_BEEP(exp)
    - FS_VERIFY(exp)
    - FS_VERIFY_MSG(exp, msg)
    - FS_VERIFY_BEEP(exp)
    - FS_STATIC_ASSERT(exp)
            
    In the <b>debug mode</b> - if the <b>expression</b> is evaluated to
    <b>false</b> then #FS_ASSERT and #FS_ASSERT_MSG invoke an <b>information
    box</b> showing information about the <b>expression and the exact code
    position</b>. In the message version a <b>user message</b> is printed as
    well. By pressing one of the buttons in the box it is possible to choose
    whether or not should be the <b>execution terminated</b>. In the case of
    #FS_ASSERT_BEEP only a <b>sound</b> is played and there is <b>no way to
    stop the execution</b> (this behaviour is especially useful in runtime
    environments when the assert is invoked within a loop and pressing a
    button every time would be annoying).

    \image html fs_sys_fsAssert_01.gif
    
    <b>FS_VERIFY_...</b> macros are used similarly, the only difference is
    the <b>expression is always evaluated!</b> (but the result is <b>not
    checked</b> in the release mode).

    The last variant <b>FS_STATIC_ASSERT</b> is used for <b>compile-time
    tests</b> e.g. <i>FS_STATIC_ASSERT(sizeof(T) == 24);</i>
        
    The main advantage of the information box is that the error information
    is available during testing where there is <b>no development
    environment</b> involved to catch the bug directly by a programmer.
    
    \par Configuration
    
    - #FS_ASSERT_LABEL
    - #FS_ASSERT_ENABLE_BOX
    - #FS_ASSERT_ENABLE_BEEP
    - \ref sc_config "/more info/"
        
    \par Implementation notes:
    
    Currently, the message & beep versions can be fully utilized on Windows
    platforms (incl. WinCE). On other platforms the standard assert
    implementation is invoked.
*/
//---------------------------------------------------------------------------
#ifndef FS_SYS_FSASSERT_H
#define FS_SYS_FSASSERT_H
//---------------------------------------------------------------------------
#if defined(FS_INCLUDE_USERDEFS) // inject user definition file on request
#   include <fs/fsUserDefs.h>
#endif 
//---------------------------------------------------------------------------
//--- EXTERNAL CONFIGURATION >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
//---------------------------------------------------------------------------
#if !defined(FS_MMGR_USERDEFS) // define to overwrite default definitions
#   if !defined(NDEBUG)
        //! Set: Information box label.
        /*! By defining this directive (a string literal), it is possible
            to customize the label of the information box. */
#       define FS_ASSERT_LABEL "FIPS' ASSERT WARNING"
        //! Enable: Standard (box) version.
        /*! If this directive is defined, it enables the message versions
            of FS_ASSERT_... / FS_VERIFY_... (typically in the debug mode).
            Otherwise the checking is disabled. Note that in the case of
            FS_VERIFY_... the expression is always evaluated! */
#       define FS_ASSERT_ENABLE_BOX
        //! Enable: Beep version.
        /*! If this directive is defined, it enables the beep versions of
            FS_ASSERT_... / FS_VERIFY_... (typically in the debug mode).
            Otherwise the checking is disabled. Note that in the case of
            FS_VERIFY_... the expression is always evaluated! */
#       define FS_ASSERT_ENABLE_BEEP
#   endif // NDEBUG
#endif // FS_MMGR_USERDEFS
//---------------------------------------------------------------------------
//--- UNDEFINE TRACKING MACROS >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
//---------------------------------------------------------------------------
#if !defined(__FUNCTION__)
#   define __FUNCTION__ "???" // MSVC feature
#endif
//---------------------------------------------------------------------------
namespace fs { namespace sys { namespace asr {
//---------------------------------------------------------------------------
#if defined(_WIN32) && defined(FS_ASSERT_ENABLE_BOX)
    //! Invokes a message box on Windows platforms (for internal use only).
    bool BoxWin(const char *, const char *, const char *, const char *, int);
#endif
//---------------------------------------------------------------------------
#if defined(_WIN32) && defined(FS_ASSERT_ENABLE_BEEP)
    //! Beeps on Windows platforms (for internal use only).
    void BeepWin();
#endif
//---------------------------------------------------------------------------
#if defined(UNDER_CE) && defined(FS_ASSERT_ENABLE_BOX)
    //! Terminates execution on the WinCE platform (for internal use only).
    void BreakWce();
#endif
//---------------------------------------------------------------------------
}}} // namespace fs::sys::asr
//---------------------------------------------------------------------------
#if defined(FS_ASSERT_ENABLE_BOX)
#   if defined(UNDER_CE)
#       define FS_ASSERT_BREAK { fs::sys::asr::BreakWce(); }
#   elif defined(_WIN32)
#       define FS_ASSERT_BREAK __asm int 3
#   else
#       define FS_ASSERT_BREAK {}
#   endif
#endif
//---------------------------------------------------------------------------
#if defined(FS_ASSERT_ENABLE_BOX)
#   if defined(_WIN32)
#       define FS_ASSERT_MSG(exp, msg)                                      \
        {                                                                   \
            bool bExp = (exp) ? true : false;                               \
            if(!bExp && fs::sys::asr::BoxWin(#exp, msg, __FUNCTION__,       \
             __FILE__, __LINE__)) FS_ASSERT_BREAK                           \
        }
#   else
#       define FS_ASSERT_MSG(exp, msg) assert(exp && msg)
#   endif
#else
#   define FS_ASSERT_MSG(exp, msg) (void(0))
#endif
//---------------------------------------------------------------------------
#define FS_ASSERT(exp) FS_ASSERT_MSG(exp, 0)
//---------------------------------------------------------------------------
#if defined(FS_ASSERT_ENABLE_BEEP)
#   if defined(_WIN32)
#       define FS_ASSERT_BEEP(exp)                                          \
        {                                                                   \
            bool bExp = (exp) ? true : false;                               \
            if(!bExp) fs::sys::asr::BeepWin();                              \
        }
#   else
#       define FS_ASSERT_BEEP(exp) (void(0))
#   endif
#else
#   define FS_ASSERT_BEEP(exp) (void(0))
#endif
//---------------------------------------------------------------------------
#if defined(FS_ASSERT_ENABLE_BOX)
#   define FS_VERIFY(exp) FS_ASSERT(exp)
#else
#   define FS_VERIFY(exp) (void(exp))
#endif
//---------------------------------------------------------------------------
#if defined(FS_ASSERT_ENABLE_BOX)
#   define FS_VERIFY_MSG(exp, msg) FS_ASSERT_MSG(exp, msg)
#else
#   define FS_VERIFY_MSG(exp, msg) (void(exp))
#endif
//---------------------------------------------------------------------------
#if defined(FS_ASSERT_ENABLE_BEEP)
#   define FS_VERIFY_BEEP(exp) FS_ASSERT_BEEP(exp)
#else
#   define FS_VERIFY_BEEP(exp) (void(exp))
#endif
//---------------------------------------------------------------------------
template<bool> struct FS_STATIC_ASSERT___E_R_R_O_R___;
//! Helper template for compile-time error checking.
template<> struct FS_STATIC_ASSERT___E_R_R_O_R___<true> {};
#define FS_STATIC_ASSERT(exp)                                               \
 ( (FS_STATIC_ASSERT___E_R_R_O_R___<(exp) != 0>()) )
//---------------------------------------------------------------------------
#endif // FS_SYS_FSASSERT_H