/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
/*                          _______      ______    ______          __      */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~    / ____(_)___ / ___ /)  / ____/___  ____/ /__    */
/*       [fsLog]          / /_  / / __ \\__ \|/  / /   / __ \/ __  / _ \   */
/*       rev.  12        / __/ / / /_/ /__/ /   / /___/ /_/ / /_/ /  __/   */
/*    29th Jan 2007     /_/   /_/ ,___/____/    \____/\____/\__,_/\___/    */
/*     [x] stable              /_/ (c) 2004-7 Filip STOKLAS (FipS)         */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~       http://HOLE.4FipS.com/fips_code.php          */
/*                                                                         */
/* This code is free for personal and commercial use. You may redistribute */
/* it by any means. If you use the code for your own projects please give  */
/* me credit. Please send a bug report. Don't alter or remove this header! */
/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
#ifndef FS_SYS_FSLOG_H
#define FS_SYS_FSLOG_H
//---------------------------------------------------------------------------
#if defined(FS_INCLUDE_USERDEFS) // inject user definition file on request
#   include <fs/fsUserDefs.h>
#endif
//---------------------------------------------------------------------------
//--- EXTERNAL CONFIGURATION >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
//---------------------------------------------------------------------------
#if !defined(FS_LOG_USERDEFS) // define to overwrite default definitions
    //! Enable: ModLog() - Module log instance.
    /*! If this directive is defined, an instance of Log_t is created for
        each module. It's accessible through fs::sys::ModLog(). Two helper
        macros #FS_LOG and #FS_LOG_DBG are also provided. */
#   define FS_LOG_ENABLE_MODLOG
    //! Set: Default configuration flags for ModLog().
#   define FS_LOG_MODLOG_MODE fs::sys::Log_t::LOG_HIST
    //! Set: Default history depth for ModLog().
#   define FS_LOG_MODLOG_HIST 1024
    //! Set: Default log file for ModLog().
#   define FS_LOG_MODLOG_FILE "fs.log"
#   if !defined(NDEBUG)
        //! Enable: #FS_LOG_DBG macro.
#       define FS_LOG_ENABLE_DEBUG
#   endif
#endif
//---------------------------------------------------------------------------
//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
//---------------------------------------------------------------------------
namespace fs { namespace sys {
//---------------------------------------------------------------------------
//! Log system.
/*! An extra effort was made to provide two key features: Individual log
    instances can be <b>linked</b> together and they can rest in
    <b>different DLL modules</b>. Before a module is connected, logged data
    are typically stored in log <b>history</b>. After connecting, the
    history is <b>dumped</b> into the main log and all future logged data
    are also <b>redirected</b>. The Fig. below shows such a module-based
    system.
    
    \image html fs_sys_Log_t_01.gif
    
    The internal logic of Log_t recognizes <b>special control sequences</b>
    in the message. Prefix '!.' means <b>warning</b>, 'x.' means
    <b>error</b>. Postfix '@' is an explicit request for <b>message
    break</b> invoking through the #FS_ASSERT_MSG macro.
    
    \par Configuration

    - #FS_LOG_ENABLE_MODLOG
    - #FS_LOG_MODLOG_MODE
    - #FS_LOG_MODLOG_HIST
    - #FS_LOG_MODLOG_FILE
    - #FS_LOG_ENABLE_DEBUG
    - \ref sc_config "/more info/"
          
    \par Implementation notes:
    
    The <b>PIMPL</b> idiom and the <b>virtual interface</b> concept are used
    together to cross <b>DLL bounds</b> if necessary.
*/
//---------------------------------------------------------------------------
class Log_t
{
 public:
 
    //! Configuration flags.
    enum Mode_e
    {
        LOG_HIST   = 0x00000001, //!< Keep records in history.
        LOG_FILE   = 0x00000002, //!< Log to a file.
        LOG_STDOUT = 0x00000004, //!< Log to the standard output.
        LOG_DBGOUT = 0x00000008, //!< Log through OutputDebugString.
        LOG_RTMBRK = 0x00000010  //!< Enable runtime break.
    };
    
    //! Initializes a new log instance using the given configuration.
    explicit Log_t(unsigned int fMode, size_t uHist = 1024,
     const char *pszFile = "fs.log");
    //! Releases links and resources (e.g. closes log file).
    virtual ~Log_t();
        
    //! Logs a formated string ('printf' style).
    void Push(const char *pszFmt, ...);
    //! Sets a new target log (dumps history, redirects future records).
    virtual void SetTgt(Log_t *pTgt);
    
 protected:
 
    friend class LogImpl_t;
    Log_t(); //!< Used only by 'LogImpl_t'.
    
    //! Links / unlinks log instances.
    virtual void RegSrc(Log_t *pSrc, bool bReg = true);
    //! Logs a text line.
    virtual void PushLn(const char *pszText, bool bMsgBrk = false);

 private:

    Log_t(const Log_t &); //!< No copy.
    Log_t & operator = (const Log_t &); //!< No assignment.
     
    Log_t *m_pImpl; //!< The PIPML idiom is used.
};
//---------------------------------------------------------------------------
#if defined (FS_LOG_ENABLE_MODLOG)
    //! Module log instance.
    Log_t & ModLog();
#endif
//---------------------------------------------------------------------------
#if defined(FS_LOG_ENABLE_MODLOG)
#   define FS_LOG fs::sys::ModLog().Push
#   if defined(FS_LOG_ENABLE_DEBUG)
#       define FS_LOG_DBG FS_LOG
#   else
#       define FS_LOG_DBG 1 ? ((void)0) : FS_LOG
#   endif
#endif
//---------------------------------------------------------------------------
}} // namespace fs::sys
//---------------------------------------------------------------------------
#endif // FS_SYS_FSLOG_H