/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
/*                          _______      ______    ______          __      */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~    / ____(_)___ / ___ /)  / ____/___  ____/ /__    */
/*      [fsMemMgr]        / /_  / / __ \\__ \|/  / /   / __ \/ __  / _ \   */
/*       rev. 24         / __/ / / /_/ /__/ /   / /___/ /_/ / /_/ /  __/   */
/*    20th Jan 2007     /_/   /_/ ,___/____/    \____/\____/\__,_/\___/    */
/*     [x] stable              /_/ (c) 2005-7 Filip STOKLAS (FipS)         */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~       http://HOLE.4FipS.com/fips_code.php          */
/*                                                                         */
/* This code is free for personal and commercial use. You may redistribute */
/* it by any means. If you use the code for your own projects please give  */
/* me credit. Please send a bug report. Don't alter or remove this header! */
/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
/*! \file fsMemMgr.h
    \brief Memory manager / enabler.
    
    fsMemMgr is a simple <b>memory manager</b> that is able to address some
    common <b>memory allocation problems</b> like <b>memory leak</b>
    detection and some <b>wrong pointer</b> handling issues. <b>Detailed
    information</b> about memory leaks is <b>dumped</b> into a
    <a href = "fs_mmgr_main.log">log file</a>.
    
    The memory manager can be easily integrated into a project. It consists
    of only 3 files: fsMemMgr.h, fsMemMgr_Not.h and fsMemMgr.cpp. The first
    one <b>enables</b> the memory manager and should be included <b>as soon
    as possible</b> (typically in a configuration header). Then all
    (de)allocations through the '<b>new</b>' and '<b>delete</b>' operators
    are <b>redirected</b> into the memory manager so that memory
    (de)allocations can be tracked from now.

    Since the memory manager uses <b>macros</b> to wrap up the '<b>new</b>'
    and '<b>delete</b>' operators it is necessary to <b>guard</b> some code
    where the operators occur in another meaning than the memory
    (de)allocation is. For more information see fsMemMgr_Not.h
    
    \par Configuration
    
    - #FS_MMGR_ENABLE_BASIC_STATS
    - #FS_MMGR_ENABLE_TRACKING
    - #FS_MMGR_ENABLE_RTM_BREAK
    - \ref sc_config "/more info/"
    
    \par Implementation notes:
    
    Since the memory manager has to store the tracking information in some
    kind of <b>container</b> and it would be nice to use a <b>standard</b>
    one (std::map) to keep things simple. It is necessary to provide a
    <b>custom STL memory allocator</b> in order to <b>prevent recursion</b>
    during internal (de)allocations. Hence the allocator uses
    <b>::malloc</b> and <b>::free</b> to remove itself from the tracking.
    <small><i>(Note that there are two versions of the allocator: the
    standard one and an old SGI's one, it enables the memory manager to be
    used on my WinCE / PocketPC 2002 platform as well).</i></small>
*/
//---------------------------------------------------------------------------
#include "fsMemMgr_NoT.h"
//---------------------------------------------------------------------------
#if !defined(__FUNCTION__)
#   define __FUNCTION__ "???" // MSVC feature
#endif // __FUNCTION__
//---------------------------------------------------------------------------
#if defined(FS_MMGR_ENABLE_TRACKING) ||                                     \
    defined(FS_MMGR_ENABLE_POOLING)  ||                                     \
    defined(FS_MMGR_ENABLE_BASIC_STATS)                                     \
    //! Defined if the internal (de)allocation operators are provided.
#   define FS_MMGR_OWN_OPERATORS
#endif
//---------------------------------------------------------------------------
//--- DEFINE TRACKING MACROS >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
//---------------------------------------------------------------------------
#if defined(FS_MMGR_ENABLE_TRACKING)
//---------------------------------------------------------------------------
#if defined(new) || defined(delete)
#   error Neither 'new' nor 'delete' should be defined here!
#endif
//---------------------------------------------------------------------------
/*! The macro wraps the 'new' operator so that the memory tracking code
    will be executed in parallel. Note that the syntax of a 'new'
    statement is not affected (in the case of memory allocation).
    The macro expands to: ptr = (Track(...), false) ? 0 : new T;
    \note The comma operator is involved to hide the tracking code.
*/
#define new (fs::sys::mmgr::aux::Track(__FUNCTION__, __FILE__,              \
 __LINE__), false) ? 0 : new
//---------------------------------------------------------------------------
/*! The macro wraps the 'delete' operator so that the memory tracking code
    will be executed in parallel. Note that the syntax of a 'delete'
    statement is not affected (in the case of memory deallocation).
    The macro expands to: (Track(...), false) ? Track(.0.) : delete ptr;
    \note The comma operator is involved to hide the tracking code. The
     second 'Track' is never called, it stands for 'void'.
*/
#define delete (fs::sys::mmgr::aux::Track(__FUNCTION__, __FILE__,           \
 __LINE__), false) ? fs::sys::mmgr::aux::Track("", "", 0) : delete
//---------------------------------------------------------------------------
#endif // FS_MMGR_ENABLE_TRACKING
//---------------------------------------------------------------------------
//--- MEMORY MANAGER STUFF >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
//---------------------------------------------------------------------------
#ifndef FS_SYS_MEMMGR_H
#define FS_SYS_MEMMGR_H
//---------------------------------------------------------------------------
namespace fs { namespace sys {
//---------------------------------------------------------------------------
//! <b>M</b>emory <b>m</b>ana<b>g</b>e<b>r</b> interface.
namespace mmgr {
//---------------------------------------------------------------------------
//! Sets a name of the current static module (useful e.g. for multiple DLLs).
void SetModNm(const char *pszModNm);
//---------------------------------------------------------------------------
#if defined(FS_MMGR_ENABLE_TRACKING)
//! Memory manager <b>aux</b>iliary interface.
namespace aux {
//! Sets a code pos. for the last memory operation (for internal use only).
void Track(const char *pszFunc, const char *pszFile, int nLine);
} // namespace aux
#endif // FS_MMGR_ENABLE_TRACKING
//---------------------------------------------------------------------------
}}} // namespace fs::sys::mmgr
//---------------------------------------------------------------------------
//--- MEMORY ALLOCATION OPERATORS >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
//---------------------------------------------------------------------------
#if defined(FS_MMGR_OWN_OPERATORS)
//---------------------------------------------------------------------------
#include "fsMemMgr_NoT.h" // turn off the tracking macros for a while
//---------------------------------------------------------------------------
//! A monitored version of the 'new' operator.
void * operator new (size_t uSize);
//! A monitored version of the 'delete' operator.
void operator delete (void *pAddr);
//! A monitored version version of the 'new []' operator.
void * operator new [] (size_t uSize);
//! A monitored version of the 'delete []' operator.
void operator delete [] (void *pAddr);
//---------------------------------------------------------------------------
#include "fsMemMgr.h" // turn on the tracking macros
//---------------------------------------------------------------------------
#endif // FS_MMGR_OWN_OPERATORS
//---------------------------------------------------------------------------
//--- <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
//---------------------------------------------------------------------------
#endif // FS_SYS_MEMMGR_H