/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
/*                          _______      ______    ______          __      */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~    / ____(_)___ / ___ /)  / ____/___  ____/ /__    */
/*      [fsStrUt]         / /_  / / __ \\__ \|/  / /   / __ \/ __  / _ \   */
/*       rev.  7         / __/ / / /_/ /__/ /   / /___/ /_/ / /_/ /  __/   */
/*    10th Jan 2007     /_/   /_/ ,___/____/    \____/\____/\__,_/\___/    */
/*     [x] stable              /_/ (c) 2006-7 Filip STOKLAS (FipS)         */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~       http://HOLE.4FipS.com/fips_code.php          */
/*                                                                         */
/* This code is free for personal and commercial use. You may redistribute */
/* it by any means. If you use the code for your own projects please give  */
/* me credit. Please send a bug report. Don't alter or remove this header! */
/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
#ifndef FS_UT_FSSTRUT_H
#define FS_UT_FSSTRUT_H
//---------------------------------------------------------------------------
#if defined(FS_INCLUDE_USERDEFS) // inject user definition file on request
#   include <fs/fsUserDefs.h>
#endif
//---------------------------------------------------------------------------
namespace fs { namespace ut {
//---------------------------------------------------------------------------
//! <b>Str</b>ing utilities.
namespace str {
//---------------------------------------------------------------------------
//! Converts an int value to a string.
char * Int2Str(int nVal, char *pszStr);
//! Converts a wide-character string to a single-byte-character string.
void WStr2Str(const wchar_t *pwszWStr, char *pszStr);
//! Converts a single-byte-character string to a wide-character string.
void Str2WStr(const char *pszStr, wchar_t *pwszWStr);
//---------------------------------------------------------------------------
//! String holder.
/*! The class especially serves as a <b>lightweight string converter /
    holder</b>. It is useful for example when an API which requires a
    <b>different string format</b> is called, so the converted string can
    be <b>temporarily stored</b> on the stack and <b>automatically
    destroyed<b> at the end of the <b>scope</b>.
*/
class StrHldr_t
{
 public:

    //! Stores the given string.
    explicit StrHldr_t(const char *pszStr);
    //! Converts from the given wide-character string and stores the result.
    explicit StrHldr_t(const wchar_t *pwszWStr);
    ~StrHldr_t();

    //! Returns the stored string (can be 0).
    const char * GetStr() const;

 private:

    StrHldr_t(); //!< No def ctor.
    StrHldr_t(const StrHldr_t &); //!< No copy.
    StrHldr_t & operator = (const StrHldr_t &); //!< No assignment.

    char *m_pszStr;
};
//---------------------------------------------------------------------------
//! Wide-character string holder.
/*! The class especially serves as a <b>lightweight string converter /
    holder</b>. It is useful for example when an API which requires a
    <b>different string format</b> is called, so the converted string can
    be <b>temporarily stored</b> on the stack and <b>automatically
    destroyed<b> at the end of the <b>scope</b>.
*/
class WStrHldr_t
{
 public:

    //! Stores the given wide-character string.
    explicit WStrHldr_t(wchar_t *pwszWStr);
    //! Converts from the given single-byte string and stores the result.
    explicit WStrHldr_t(const char *pszStr);
    ~WStrHldr_t();

    //! Returns the stored wide-character string (can be 0).
    const wchar_t * GetWStr() const;

 private:

    WStrHldr_t(); //!< No def ctor.
    WStrHldr_t(const WStrHldr_t &); //!< No copy.
    WStrHldr_t & operator = (const WStrHldr_t &); //!< No assignment.

    wchar_t *m_pwszWStr;
};
//---------------------------------------------------------------------------
//! String tokenizer.
/*! Toknzr_t <b>parses</b> the <b>input text</b> sequentially and provides
    individual <b>tokens</b> according to the given <b>criteria</b>
    (<b>quoted tokens</b> are also supported). Note that Toknzr_t
    <b>doesn't copy</b> the <b>input text</b>, so there is <b>no memory
    overhead</b>. The example of the usage see below.
    
    \code
    fs::ut::str::Toknzr_t Toknzr("one, two, \"t h r e e\"", " ,", '\"');
    for(const char *pszTok = Toknzr.Next(); pszTok; pszTok = Toknzr.Next())
    {
        ::printf("token: %s\n", pszTok);
    }
    \endcode
*/
class Toknzr_t
{
 public:
 
    //! Receives the input text and parsing criteria.
    explicit Toknzr_t(const char *pszText, const char *pszSprs = " \t",
     char chQuote = '\"');
    ~Toknzr_t();
     
    //! Returns the next token (0 - no more tokens or error).
    const char * Next();
 
 private:
 
     Toknzr_t(); //!< No def ctor.
     Toknzr_t(const Toknzr_t &); //!< No copy.
     Toknzr_t & operator = (const Toknzr_t &); //!< No assignment.

     void *m_pImpl; //!< The PIPML idiom is used.
};
//---------------------------------------------------------------------------
//! Command line parser.
/*! The class <b>parses</b> the given <b>command line</b> and manages
    <b>tokens</b> and <b>key / value</b> pairs. The expected command line
    <b>format</b> is: <i>'foo -key1 value1 -switch -key2 "complex value2"
    ...'</i>. Later, it is possible to <b>query</b> the <b>value</b> of a
    <b>key</b> using #GetVal (both are strings). The <b>keys</b> are
    <b>prefixed</b> with '<b>-</b>' symbol. Note that all consecutive keys
    except for the one just before the value are discarded. Similarly, if
    there are consecutive values after the key, only the first value is
    employed. Quotes can be used to form more complex values.
*/
class CmdLn_t
{
 public:
 
    //! Parses the given command line string.
    explicit CmdLn_t(const char *pszCmdLn);
    //! Parses the standard command line.
    CmdLn_t(int argc, char *argv[]);
    ~CmdLn_t();
    
    //! Returns a value for the given key (0 if not found).
    const char * GetVal(const char *pszKey) const;
    
    //! Returns the total number of tokens.
    int GetNumToks() const;
    //! Returns a token at the given index (0 if out of range).
    const char * GetTok(int nIdx) const;
    //! Finds the index of the token (-1 if not found).
    int FindTok(const char *pszTok) const;
                     
 private:
 
    CmdLn_t(); //!< No def ctor.
    CmdLn_t(const CmdLn_t &); //!< No copy.
    CmdLn_t & operator = (const CmdLn_t &); //!< No assignment.

    void *m_pImpl; //!< The PIPML idiom is used.
};
//---------------------------------------------------------------------------
}}} // namespace fs::ut::str
//---------------------------------------------------------------------------
#endif // FS_UT_FSSTRUT_H