/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
/*                          _______      ______    ______          __      */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~    / ____(_)___ / ___ /)  / ____/___  ____/ /__    */
/*     [fsXmlDoc]         / /_  / / __ \\__ \|/  / /   / __ \/ __  / _ \   */
/*       rev. 4          / __/ / / /_/ /__/ /   / /___/ /_/ / /_/ /  __/   */
/*    13th Jan 2007     /_/   /_/ ,___/____/    \____/\____/\__,_/\___/    */
/*     [x] stable              /_/ (c) 2005-7 Filip STOKLAS (FipS)         */
/*  ~ ~ ~ ~ ~ ~ ~ ~ ~ ~       http://HOLE.4FipS.com/fips_code.php          */
/*                                                                         */
/* This code is free for personal and commercial use. You may redistribute */
/* it by any means. If you use the code for your own projects please give  */
/* me credit. Please send a bug report. Don't alter or remove this header! */
/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
#include <fs/ut/fsXmlDoc.h>
#include <fs/ut/fsStrUt.h> // tokenizer

// >>> COMMON SUPPORT BLOCK (rev. 6)
#if defined(FS_HAS_FSASSERT)
#   include <fs/sys/fsAssert.h>
#else
#   include <cassert>
#   if !defined(FS_ASSERT)
#       define FS_ASSERT(exp) assert(exp)
#   endif
#   if !defined(FS_ASSERT_MSG)
#       define FS_ASSERT_MSG(exp, msg) assert(exp && msg)
#   endif
#   if !defined(FS_VERIFY)
#       define FS_VERIFY(exp) \
        { bool bExp = !!(exp); assert(bExp && #exp); bExp; }
#   endif
#   if !defined(FS_VERIFY_MSG)
#       define FS_VERIFY_MSG(exp, msg) \
        { bool bExp = !!(exp); assert(bExp && #exp && msg); bExp; }
#   endif
#   if !defined(FS_STATIC_ASSERT)
#       define FS_STATIC_ASSERT(exp) { char error[(exp) ? 1 : 0]; error; }
#   endif
#endif
#if !defined(FS_VERIFY_RETURN)
#   define FS_VERIFY_RETURN(exp) \
    if(!(exp)) { FS_ASSERT_MSG(0, #exp "<return>"); return; }
#endif
#if !defined(FS_VERIFY_RETURN_VAL)
#   define FS_VERIFY_RETURN_VAL(exp, ret) \
    if(!(exp)) { FS_ASSERT_MSG(0, #exp "<returns:>" #ret); return (ret); }
#endif
// <<< COMMON SUPPORT BLOCK (rev. 6)

#include <string>
#include <memory> // auto_ptr

// provide detailed memory tracking information
#if defined(FS_HAS_FSMEMMGR)
#   include <fs/sys/fsMemMgr.h>
#endif
//---------------------------------------------------------------------------
#include <tinyxml/tinyxml.h>
//---------------------------------------------------------------------------
using namespace fs::ut;
//---------------------------------------------------------------------------
//--- ANONYMOUS NAMESPACE >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
//---------------------------------------------------------------------------
namespace {
//---------------------------------------------------------------------------
//--- CLASS XmlDocImpl_t >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
//---------------------------------------------------------------------------
class XmlDocImpl_t
{
 public:
 
    explicit XmlDocImpl_t(const char *pszFile);
    XmlDocImpl_t(const char *pszData, bool bUnused);
    bool IsValid() const;

    bool Query(const char *pszQuery, int &nValue) const;
    bool Query(const char *pszQuery, const char *&pszText) const;
    
 private:
    
    XmlDocImpl_t();
    XmlDocImpl_t(const XmlDocImpl_t &);
    XmlDocImpl_t & operator = (const XmlDocImpl_t &);
    
    //! Returns the number of children of the element with the given label.
    int CountChdrn(const TiXmlElement *pElem, const char *pszLabel) const;
    //! Traverses the tree according to the query, returns the leaf.
    const TiXmlElement * TravToLeaf(const char *pszQuery, std::string
     &strLeafReq) const;
    
    //---
    
    typedef std::auto_ptr<TiXmlDocument> DocPtr_t;
    DocPtr_t m_ptrDoc;
};
//---------------------------------------------------------------------------
XmlDocImpl_t::XmlDocImpl_t(const char *pszFile)
{
    FS_VERIFY_RETURN(pszFile);

    DocPtr_t ptrDoc = DocPtr_t(new TiXmlDocument);
    FS_VERIFY_RETURN(ptrDoc->LoadFile(pszFile));
    FS_VERIFY_RETURN(!ptrDoc->Error());
            
    //--- TRANSFER THE STATE ---
    
    m_ptrDoc = ptrDoc;
}
//---------------------------------------------------------------------------
XmlDocImpl_t::XmlDocImpl_t(const char *pszData, bool /*bUnused*/)
{
    FS_VERIFY_RETURN(pszData);

    DocPtr_t ptrDoc = DocPtr_t(new TiXmlDocument);
    ptrDoc->Parse(pszData);
    FS_VERIFY_RETURN(!ptrDoc->Error());
            
    //--- TRANSFER THE STATE ---
    
    m_ptrDoc = ptrDoc;
}
//---------------------------------------------------------------------------
bool XmlDocImpl_t::IsValid() const
{
    return m_ptrDoc.get() ? true : false;
}
//---------------------------------------------------------------------------
bool XmlDocImpl_t::Query(const char *pszQuery, int &nValue) const
{
    FS_VERIFY_RETURN_VAL(pszQuery, false);

    std::string strLeafReq;
    const TiXmlElement *pLeaf = TravToLeaf(pszQuery, strLeafReq);
    
    if(!pLeaf) // non-existent XML path
        return false;

    if(strLeafReq.empty()) // return the content
    {
        const char *pszValue = pLeaf->GetText();
        if(pszValue)
        {
            nValue = ::atoi(pszValue);
            return true;
        }
    }
    else if(strLeafReq[0] == '#') // count children
    {
        FS_VERIFY_RETURN_VAL(strLeafReq.size() > 1, false);
        nValue = CountChdrn(pLeaf, strLeafReq.c_str() + 1);
        return true;
    }
    else if(strLeafReq[0] == '@') // get the attribute
    {
        FS_VERIFY_RETURN_VAL(strLeafReq.size() > 1, false);
        int nTmp;
        if(pLeaf->Attribute(strLeafReq.c_str() + 1, &nTmp))
        {
            nValue = nTmp;
            return true;
        }
    }

    return false;
}
//---------------------------------------------------------------------------
bool XmlDocImpl_t::Query(const char *pszQuery, const char *&pszText) const
{
    FS_VERIFY_RETURN_VAL(pszQuery, false);

    std::string strLeafReq;
    const TiXmlElement *pLeaf = TravToLeaf(pszQuery, strLeafReq);

    if(!pLeaf) // non-existent XML path
        return false;
        
    if(strLeafReq.empty()) // return the content
    {
        const char *pszTmp = pLeaf->GetText();
        if(pszTmp)
        {
            pszText = pszTmp;
            return true;
        }
    }
    else if(strLeafReq[0] == '#') // count children
    {
        FS_ASSERT_MSG(0, "Not implemented / required!");
        return false;
    }
    else if(strLeafReq[0] == '@') // get the attribute
    {
        FS_VERIFY_RETURN_VAL(strLeafReq.size() > 1, false);
        const char *pszTmp = pLeaf->Attribute(strLeafReq.c_str() + 1);
        if(pszTmp)
        {
            pszText = pszTmp;
            return true;
        }
    }

    return false;
}
//---------------------------------------------------------------------------
int XmlDocImpl_t::CountChdrn(const TiXmlElement *pElem, const char *pszLabel)
 const
{
    FS_ASSERT(pElem && pszLabel);

    pElem = pElem->FirstChildElement(pszLabel);

    int n = 0;
    while(pElem)
    {
        pElem = pElem->NextSiblingElement(pszLabel);
        ++n;
    }

    return n;
}
//---------------------------------------------------------------------------
const TiXmlElement * XmlDocImpl_t::TravToLeaf(const char *pszQuery,
 std::string &strLeafReq) const
{
    FS_ASSERT(pszQuery);

    const TiXmlElement *pElem = 0;
    std::string strTokTmp;
        
    fs::ut::str::Toknzr_t Toknzr(pszQuery, "/");
    for(const char *pszTok = Toknzr.Next(); pszTok; pszTok = Toknzr.Next())
    {
        // look for the branching symbol '^'
        const char *pszBch = ::strchr(pszTok, '^');
        
        // '#' and '@' are always treated as a leaf
        if(*pszTok == '@' || *pszTok == '#')
        {
            FS_VERIFY_RETURN_VAL(!pszBch, 0); // unexpected '^'
            FS_VERIFY_RETURN_VAL(pElem, 0); // '@/#' invalid for the root
            
            strLeafReq = pszTok;
            return pElem;
        }
                        
        // get the branching index, crop the token
        int nBchIdx = 0;
        if(pszBch)
        {
            FS_VERIFY_RETURN_VAL(pElem, 0); // '^' invalid for the root
        
            // get the branching index
            nBchIdx = ::atoi(pszBch + 1);
            // crop the branching information
            strTokTmp = std::string(pszTok, pszBch - pszTok);
            pszTok = strTokTmp.c_str(); // substitute
        }
                                                                
        // go down the tree (use the doc or the previous element)
        pElem = !pElem ? m_ptrDoc->FirstChildElement(pszTok)
         : pElem->FirstChildElement(pszTok);
        
        if(!pElem) // requested element doesn't exist
            return 0;
                                        
        // switch to the requested branch (on the same level)
        while(nBchIdx > 0)
        {
            pElem = pElem->NextSiblingElement(pszTok);

            if(!pElem) // out of range
                return 0;
            else --nBchIdx;
        }
    }
    
    return pElem; // bare leaf (strLeafReq is empty)
}
//---------------------------------------------------------------------------
} // anonymous namespace
//---------------------------------------------------------------------------
//--- CLASS XmlDoc_t >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
//---------------------------------------------------------------------------
XmlDoc_t::XmlDoc_t(const char *pszFile):
m_pImpl(new XmlDocImpl_t(pszFile))
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
}
//---------------------------------------------------------------------------
XmlDoc_t::XmlDoc_t(const char *pszData, bool bUnused):
m_pImpl(new XmlDocImpl_t(pszData, bUnused))
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
}
//---------------------------------------------------------------------------
XmlDoc_t::~XmlDoc_t()
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
    delete static_cast<XmlDocImpl_t *>(m_pImpl);
}
//---------------------------------------------------------------------------
bool XmlDoc_t::IsValid() const
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
    return static_cast<XmlDocImpl_t *>(m_pImpl)->IsValid();
}
//---------------------------------------------------------------------------
/*! \param pszQuery Query string consisting of a path and a request.
    \param nValue Reference to a variable receiving the resulting value.
     Note that if the query fail, nValue is not affected.
    \return <b>true</b> indicates success.
*/
bool XmlDoc_t::Query(const char *pszQuery, int &nValue) const
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
    return static_cast<XmlDocImpl_t *>(m_pImpl)->Query(pszQuery, nValue);
}
//---------------------------------------------------------------------------
/*! \param pszQuery Query string consisting of a path and a request.
    \param pszText Reference to a variable receiving the resulting text.
     Note that if the query fail, pszText is not affected.
    \return <b>true</b> indicates success.
*/
bool XmlDoc_t::Query(const char *pszQuery, const char *&pszText) const
{
    FS_ASSERT_MSG(m_pImpl, "Invalid implementation pointer!");
    return static_cast<XmlDocImpl_t *>(m_pImpl)->Query(pszQuery, pszText);
}
//---------------------------------------------------------------------------